# -*- coding: utf-8 -*-
'''This module implements a BackupPC configuration generator'''

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Authors:
#  * Stéphane Graber <stgraber@ubuntu.com>, 2009-2010 Revolution Linux
#  * Vincent Vinet <vincent.vinet@revolutionlinux.com>, 2009-2010 Revolution Linux
#  * Fernando Laudares Camargos <fernando@revolutionlinux.com>, 2009-2010 Revolution Linux

# Code based on the NagiosGen plugin

import re, os, glob, socket, logging
import cStringIO
import configobj
import Bcfg2.Server.Plugin

LOGGER = logging.getLogger('Bcfg2.Plugins.BackupPCGen')
PROBE_KEY = "BackupPCprobe"
PROBE_COL_DELIM = ";"
PROBE_PATH_DELIM = " "
HOSTS_DIR="hosts"
HOSTS_SECTION="hosts"

class BackupPCGen(Bcfg2.Server.Plugin.Plugin, Bcfg2.Server.Plugin.Generator):
    '''BackupPCGen is a Bcfg2 plugin that dynamically generates
       BackupPC configuration file based on Bcfg2 data.'''
    name = 'BackupPCGen'
    __version__ = '0.1'
    __author__ = 'vv@rlnx.com'

    def __init__(self, core, datastore):
        Bcfg2.Server.Plugin.Plugin.__init__(self, core, datastore)
        self.Entries = {
            'Path':
                {
                    '/etc/backuppc/backuppcgen.conf': self.generateBPCGenConf,
                    '/etc/rsyncd.conf':self.generateRsyncdConf,
                },
        }
        if not os.path.exists(os.path.join(self.data, HOSTS_DIR)):
            os.makedirs(os.path.join(self.data, HOSTS_DIR),mode=0755)

        #Used for magic
        self.client_attrib = {'encoding':'ascii',
                              'owner':'root',
                              'group':'root',
                              'type':'file',
                              'perms':'0400'}

    def generateBPCGenConf(self, entry, metadata):
        '''Build host specific configuration file'''
        #Get the backuped locations
        config = self._getServersConf()
        if not metadata.hostname in config:
            entry.text = "# BackupPC server not in servers.conf\n"
            [entry.attrib.__setitem__(key, value) for \
                (key, value) in self.client_attrib.iteritems()]
            return

        if not 'locations' in config[metadata.hostname]:
            entry.text = "# No locations defined for server\n"
            [entry.attrib.__setitem__(key, value) for \
                (key, value) in self.client_attrib.iteritems()]
            return

        locations = config[metadata.hostname]['locations'].strip().split(',')

        cstr = cStringIO.StringIO()
        config = configobj.ConfigObj()
        hostsconfig={}
        for location in locations:
            hostsconfig.update(self._getHostsConfig(location).dict())
        config[HOSTS_SECTION] = hostsconfig
        config.write(cstr)
        entry.text = cstr.getvalue()

        #Do the magic attribute stuff
        [entry.attrib.__setitem__(key, value) for \
            (key, value) in self.client_attrib.iteritems()]

    def generateRsyncdConf(self, entry, metadata):
        #Parse the probe metadata
        probe_res = metadata.Probes.get(PROBE_KEY, "")
        lines = probe_res.splitlines()
        if len(lines) < 3:
            #There is no data, abort
            LOGGER.error("No %s probe data received for host %s", PROBE_KEY, metadata.hostname)
            return
        location = lines.pop(0)
        if not location.lower().startswith("#location="):
            LOGGER.error("Malformed or missing location line: %s", location)
            return

        location = location.strip().split("=",1)[1]

        #Generate the config
        shares = configobj.ConfigObj()
        nobackup=set()
        try:
            for line in lines:
                share_name, path, excludes, includes, backup = line.strip().split(PROBE_COL_DELIM)

                shares[share_name] = {}
                if backup == "nobackup":
                    nobackup.add(share_name)
                if excludes:
                    shares[share_name]['exclude'] = excludes
                if includes:
                    shares[share_name]['include'] = includes
                shares[share_name]['path'] = path
        except ValueError:
            #Format error probably, log an error and abort
            LOGGER.error("Parsing error for %s probe data for host %s", PROBE_KEY, metadata.hostname)
            return

        #We need the rsyncd.conf template first
        template = self._getRsyncdTemplate()

        #We also need servers.conf to find out the allowed_hosts
        allowed_hosts = []
        sconf = self._getServersConf()
        for server, values in sconf.items():
            try:
                server=socket.gethostbyname(server)
            except:
                pass
            allowed_hosts.append(server)

        cstr = cStringIO.StringIO()
        cstr.write(template)
        cstr.write("hosts allow = "+" ".join(map(str,allowed_hosts)))
        cstr.write("\n\n")

        shares.write(cstr)
        entry.text = cstr.getvalue()

        #Do the magic attribute stuff
        [entry.attrib.__setitem__(key, value) for \
            (key, value) in self.client_attrib.iteritems()]

        #Write hosts/hostname.conf
        hostname_conf = configobj.ConfigObj({metadata.hostname : dict(
            location = location,
            share = [ k for k in shares if not k in nobackup ],)}
        )
        hostname_file = file(os.path.join(self.data, HOSTS_DIR, metadata.hostname + ".conf"), 'w')
        hostname_conf.write(hostname_file)
        hostname_file.close()

    def _getServersConf(self):
        """ Utility function to get the server locations list"""
        filepath = os.path.join(self.data, "servers.conf")
        if os.path.isfile(filepath):
            config = configobj.ConfigObj(filepath)
        else:
            LOGGER.error("Unable to find servers.conf")
            config = configobj.ConfigObj()
        return config

    def _getRsyncdTemplate(self):
        if not os.path.isfile(os.path.join(self.data, "rsyncd.conf.tmpl")):
            res = '# Missing rsyncd.conf.tmpl, broken rsync config\n\n'
            LOGGER.error("Unable to find rsyncd.conf.tmpl")
        else:
            filepath = os.path.join(self.data, "rsyncd.conf.tmpl")
            f = file(filepath, 'r')
            res = f.read()
            f.close()
        return res

    def _getHostsConfig(self, location):
        conf = configobj.ConfigObj()
        for fname in glob.glob(os.path.join(self.data, HOSTS_DIR, "*.conf")):
            host_conf = configobj.ConfigObj(fname)
            #Only update with matching locations
            for key, section in host_conf.items():
                if section.get("location", None) == location:
                    conf[key] = section
        return conf
